use std::collections::hash_map::Entry;
use std::collections::{HashMap, HashSet};
use std::path::Path;

use color_eyre::eyre::{eyre, ContextCompat, WrapErr};
use color_eyre::Result;
use resvg::usvg::{TextRendering, TreeParsing, TreeTextToPath};
use tiny_skia::{Pixmap, Transform};

use crate::model::config::{Config, IconFormat, IconPack};
use crate::model::icon_descriptor::{IconDescriptor, IconDescriptorSource};
use crate::model::image_filter::ImageFilterDestructive;
use crate::model::{key_page, knob_page};

mod destructive_filter;

pub type LoadedIconsMap = HashMap<(IconDescriptorSource, ImageFilterDestructive), LoadedIcon>;

#[derive(Debug)]
pub struct LoadedIcon {
    pub pixmap: Pixmap,
    pub pre_scale: f32,
}

pub fn get_used_icon_descriptors(config: &Config) -> HashSet<IconDescriptor> {
    let mut result: HashSet<IconDescriptor> = HashSet::new();

    fn insert_all_from_key_style_by_state_map<T>(result: &mut HashSet<IconDescriptor>, map: &key_page::StyleByStateMap<T>) {
        map.values().for_each(|v| {
            if let Some(icon) = &v.icon {
                result.insert(icon.clone());
            }
        });
    }

    fn insert_all_from_knob_style_by_state_map<T>(result: &mut HashSet<IconDescriptor>, map: &knob_page::StyleByStateMap<T>) {
        map.values().for_each(|v| {
            if let Some(icon) = &v.icon {
                result.insert(icon.clone());
            }
        });
    }

    for page in config.key_pages_by_id.values() {
        for key in page.keys.values() {
            if let Some(d) = &key.base_style.icon {
                result.insert(d.clone());
            }

            if let Some(c) = &key.mode.media__next {
                insert_all_from_key_style_by_state_map(&mut result, &c.style);
            }

            if let Some(c) = &key.mode.media__play_pause {
                insert_all_from_key_style_by_state_map(&mut result, &c.style);
            }

            if let Some(c) = &key.mode.media__previous {
                insert_all_from_key_style_by_state_map(&mut result, &c.style);
            }

            if let Some(c) = &key.mode.home_assistant__button {
                insert_all_from_key_style_by_state_map(&mut result, &c.style);
            }

            if let Some(c) = &key.mode.home_assistant__switch {
                insert_all_from_key_style_by_state_map(&mut result, &c.style);
            }

            if let Some(c) = &key.mode.spotify__repeat {
                insert_all_from_key_style_by_state_map(&mut result, &c.style);
            }

            if let Some(c) = &key.mode.spotify__shuffle {
                insert_all_from_key_style_by_state_map(&mut result, &c.style);
            }
        }
    }

    for page in config.knob_pages_by_id.values() {
        for knob in page.knobs.values() {
            if let Some(d) = &knob.base_style.icon {
                result.insert(d.clone());
            }

            if let Some(c) = &knob.mode.audio_volume {
                insert_all_from_knob_style_by_state_map(&mut result, &c.style)
            }
        }
    }

    result
}

pub fn load_icons(
    config_directory: &Path,
    icon_packs_by_id: &HashMap<String, IconPack>,
    descriptors: HashSet<IconDescriptor>,
    dpi: f32,
) -> Result<LoadedIconsMap> {
    let mut highest_scale_by_source: HashMap<IconDescriptorSource, f32> = HashMap::new();

    for d in &descriptors {
        let mut scale = d.filter.transform.scale;
        if let IconDescriptorSource::IconPack { pack_id, .. } = &d.source {
            let pack = &icon_packs_by_id[pack_id];
            if let Some(filter) = &pack.global_filter {
                scale *= filter.transform.scale;
            }
        }

        if let Some(v) = highest_scale_by_source.get_mut(&d.source) {
            if *v < scale {
                *v = scale;
            }
        } else {
            highest_scale_by_source.insert(d.source.clone(), scale);
        }
    }

    let mut unfiltered_pixmap_and_scale_by_source: HashMap<IconDescriptorSource, (Pixmap, f32)> = HashMap::new();
    let mut icons: LoadedIconsMap = HashMap::new();
    let mut fonts_db = resvg::usvg::fontdb::Database::new();
    fonts_db.load_system_fonts();

    for descriptor in descriptors {
        let icon_pack = if let IconDescriptorSource::IconPack { pack_id, .. } = &descriptor.source {
            Some(icon_packs_by_id.get(pack_id).wrap_err_with(|| format!("Unknown icon pack: @{}", pack_id))?)
        } else {
            None
        };

        let filter = if let Some(global_filter) = icon_pack.and_then(|p| p.global_filter.as_ref()) {
            descriptor.filter.destructive.merge_over(&global_filter.destructive)
        } else {
            descriptor.filter.destructive
        };

        let id = (descriptor.source, filter);

        if icons.contains_key(&id) {
            continue;
        }

        let (original_image, original_image_scale) = match unfiltered_pixmap_and_scale_by_source.entry(id.0.clone()) {
            Entry::Occupied(o) => o.into_mut(),
            Entry::Vacant(v) => v.insert(read_image_and_get_scale(
                config_directory,
                dpi,
                &fonts_db,
                &id.0,
                icon_pack,
                highest_scale_by_source[&id.0],
            )?),
        };

        let pixmap = destructive_filter::apply(original_image, &id.1)?;

        icons.insert(
            id,
            LoadedIcon {
                pixmap,
                pre_scale: *original_image_scale,
            },
        );
    }

    Ok(icons)
}

fn read_image_and_get_scale(
    config_directory: &Path,
    dpi: f32,
    fonts_db: &resvg::usvg::fontdb::Database,
    source: &IconDescriptorSource,
    icon_pack: Option<&IconPack>,
    highest_scale: f32,
) -> Result<(Pixmap, f32)> {
    let path = match source {
        IconDescriptorSource::Path(path) => path.clone(),
        IconDescriptorSource::IconPack { icon_id, .. } => {
            let icon_pack = icon_pack.unwrap();
            let extension = match icon_pack.format {
                IconFormat::Png => "png",
                IconFormat::Svg => "svg",
            };

            config_directory.join(&icon_pack.path).join(icon_id.to_owned() + "." + extension)
        }
    };

    Ok(match path.extension() {
        None => return Err(eyre!("Invalid icon path: {:?}", path)),
        Some(extension) => match extension.to_string_lossy().as_ref() {
            "png" => (
                Pixmap::load_png(&path).wrap_err_with(|| format!("Failed to open or decode the PNG file at {}", path.to_string_lossy()))?,
                1.0,
            ),
            "svg" => (
                read_image_from_svg(&path, dpi, fonts_db, highest_scale)
                    .wrap_err_with(|| format!("Failed to open or decode the SVG file at {}", path.to_string_lossy()))?,
                highest_scale,
            ),
            extension => return Err(eyre!("Invalid file extension, only *.png and *.svg are allowed: {}", extension)),
        },
    })
}

fn read_image_from_svg(path: &Path, dpi: f32, font_db: &resvg::usvg::fontdb::Database, scale: f32) -> Result<Pixmap> {
    let raw_data = std::fs::read(path)?;

    let tree = {
        let mut tree = resvg::usvg::Tree::from_data(
            &raw_data,
            &resvg::usvg::Options {
                dpi,
                font_family: "Inter".to_owned(),
                font_size: 11.0,
                text_rendering: TextRendering::OptimizeLegibility,
                ..resvg::usvg::Options::default()
            },
        )?;

        tree.convert_text(font_db);

        resvg::Tree::from_usvg(&tree)
    };

    let size = tree.size.to_int_size();
    let mut pixmap = Pixmap::new((size.width() as f32 * scale).ceil() as u32, (size.height() as f32 * scale).ceil() as u32).unwrap();

    tree.render(Transform::from_scale(scale, scale), &mut pixmap.as_mut());

    Ok(pixmap)
}
