use std::fmt::{Display, Formatter};
use std::str::FromStr;

use enum_map::Enum;
use serde::{Deserialize, Serialize};
use serde_with::{DeserializeFromStr, SerializeDisplay};
use thiserror::Error;

use loupedeck_serial::characteristics::LoupedeckButton;

#[derive(Debug, Eq, PartialEq, Hash, Clone, Copy, SerializeDisplay, DeserializeFromStr)]
/// One-based coordinates of a specific virtual (not physical) key.
pub struct KeyPosition {
    pub x: u16,
    pub y: u16,
}

#[derive(Debug, Error)]
#[error("The input value does not match the required format of <x> and <y> separated by an 'x'")]
pub struct KeyPositionFromStrError {}

impl FromStr for KeyPosition {
    type Err = KeyPositionFromStrError;

    fn from_str(s: &str) -> Result<Self, Self::Err> {
        let values = s.split_once('x');

        if let Some((x, y)) = values {
            if let Ok(x) = u16::from_str(x) {
                if let Ok(y) = u16::from_str(y) {
                    return Ok(KeyPosition { x, y });
                }
            }
        }

        Err(KeyPositionFromStrError {})
    }
}

impl Display for KeyPosition {
    fn fmt(&self, f: &mut Formatter<'_>) -> std::fmt::Result {
        f.write_fmt(format_args!("{}x{}", self.x, self.y))
    }
}

#[derive(Debug, Eq, PartialEq, Hash, Clone, Serialize, Deserialize)]
pub struct KeyPath {
    pub page_id: String,
    pub position: KeyPosition,
}

impl Display for KeyPath {
    fn fmt(&self, f: &mut Formatter<'_>) -> std::fmt::Result {
        f.write_fmt(format_args!("{}/{}", &self.page_id, &self.position))
    }
}

#[derive(Debug, Eq, PartialEq, Hash, Clone, Copy, Serialize, Deserialize, Enum)]
#[serde(rename_all = "kebab-case")]
pub enum KnobPosition {
    LeftTop,
    LeftMiddle,
    LeftBottom,
    RightTop,
    RightMiddle,
    RightBottom,
}

#[derive(Debug, Eq, PartialEq, Hash, Clone, Serialize, Deserialize)]
pub struct KnobPath {
    pub page_id: String,
    pub position: KnobPosition,
}

#[derive(Debug, Eq, PartialEq, Hash, Clone, Copy, Deserialize, Enum)]
pub enum ButtonPosition {
    #[serde(rename = "0")]
    N0,
    #[serde(rename = "1")]
    N1,
    #[serde(rename = "2")]
    N2,
    #[serde(rename = "3")]
    N3,
    #[serde(rename = "4")]
    N4,
    #[serde(rename = "5")]
    N5,
    #[serde(rename = "6")]
    N6,
    #[serde(rename = "7")]
    N7,
}

impl ButtonPosition {
    pub fn of(button: &LoupedeckButton) -> Self {
        match button {
            LoupedeckButton::N0 => Self::N0,
            LoupedeckButton::N1 => Self::N1,
            LoupedeckButton::N2 => Self::N2,
            LoupedeckButton::N3 => Self::N3,
            LoupedeckButton::N4 => Self::N4,
            LoupedeckButton::N5 => Self::N5,
            LoupedeckButton::N6 => Self::N6,
            LoupedeckButton::N7 => Self::N7,
        }
    }
}

impl Display for ButtonPosition {
    fn fmt(&self, f: &mut Formatter<'_>) -> std::fmt::Result {
        f.write_str(match self {
            ButtonPosition::N0 => "0",
            ButtonPosition::N1 => "1",
            ButtonPosition::N2 => "2",
            ButtonPosition::N3 => "3",
            ButtonPosition::N4 => "4",
            ButtonPosition::N5 => "5",
            ButtonPosition::N6 => "6",
            ButtonPosition::N7 => "7",
        })
    }
}
