use std::sync::Arc;

use flume::Sender;
use serde::Deserialize;
use tokio::sync::broadcast::Receiver;

use crate::model::key_page::StyleByStateMap;
use crate::model::position::KeyPath;
use crate::modes::key::KeyEvent;
use crate::runner::state::StateChangeCommand;

#[derive(Debug, Deserialize)]
pub struct PlayPauseConfig {
    #[serde(default)]
    pub style: StyleByStateMap<PlayPauseState>,
    #[serde(default)]
    pub action: PlayPauseAction,
}

#[derive(Debug, Default, Eq, PartialEq, Deserialize)]
#[serde(rename_all = "kebab-case")]
pub enum PlayPauseAction {
    #[default]
    Toggle,
    Play,
    Pause,
}

#[derive(Debug, Deserialize)]
pub struct PreviousAndNextConfig {
    #[serde(default)]
    pub style: StyleByStateMap<PreviousAndNextState>,
}

#[derive(Debug, Clone, Eq, PartialEq, Hash, Deserialize)]
#[serde(rename_all = "kebab-case")]
pub enum PlayPauseState {
    Inactive,
    Playing,
    Paused,
}

#[derive(Debug, Eq, PartialEq, Hash, Deserialize)]
#[serde(rename_all = "kebab-case")]
pub enum PreviousAndNextState {
    Inactive,
    Active,
}

pub async fn handle(path: KeyPath, config: Arc<PlayPauseConfig>, mut events: Receiver<KeyEvent>, commands: Sender<StateChangeCommand>) {
    let mut state = PlayPauseState::Inactive;

    while let Ok(event) = events.recv().await {
        if event == KeyEvent::Press {
            state = match state {
                PlayPauseState::Playing => PlayPauseState::Paused,
                _ => PlayPauseState::Playing,
            };

            commands
                .send(StateChangeCommand::SetKeyStyle {
                    path: path.clone(),
                    value: config.style.get(&state).cloned(),
                })
                .unwrap();
        }
    }
}
