use std::cell::RefCell;
use std::collections::HashMap;

use bytes::{BufMut, Bytes, BytesMut};
use resvg::usvg::tiny_skia_path::PathBuilder;
use tiny_skia::{
    BlendMode, Color, FilterQuality, IntSize, LineCap, LineJoin, Paint, Pixmap, PixmapPaint, PremultipliedColorU8, Rect, Shader, Stroke, Transform,
};

use loupedeck_serial::util::Endianness;

use crate::icons::LoadedIconsMap;
use crate::model::image_filter::ImageFilter;
use crate::runner::graphics::labels::LabelRenderer;
use crate::runner::state::Key;

#[derive(Debug)]
pub struct GraphicsContext {
    pub label_renderer: RefCell<LabelRenderer>,
    pub buffer_endianness: Endianness,
    pub global_icon_filter_by_pack_id: HashMap<String, ImageFilter>,
    pub loaded_icons: LoadedIconsMap,
}

pub fn render_key(context: &GraphicsContext, key_size: IntSize, state: Option<&Key>) -> Bytes {
    let mut pixmap = Pixmap::new(key_size.width(), key_size.height()).unwrap();

    if let Some(state) = state {
        let style = state.style.as_ref().map(|s| s.merge_over(&state.base_style));
        let style = style.as_ref().unwrap_or(&state.base_style);

        if let Some(icon) = &style.icon {
            let filter = if let Some(global_filter) = icon.source.pack_id().and_then(|i| context.global_icon_filter_by_pack_id.get(i)) {
                icon.filter.merge_over(global_filter)
            } else {
                icon.filter.clone()
            };

            let loaded_icon = &context.loaded_icons[&(icon.source.clone(), filter.destructive)];

            let scale = filter.transform.scale / loaded_icon.pre_scale;

            let scaled_size = IntSize::from_wh(loaded_icon.pixmap.width(), loaded_icon.pixmap.height())
                .unwrap()
                .scale_by(scale)
                .unwrap();

            pixmap.draw_pixmap(
                (((key_size.width() as i32 - scaled_size.width() as i32) / 2) as f32 / scale).round() as i32,
                (((key_size.height() as i32 - scaled_size.height() as i32) / 2) as f32 / scale).round() as i32,
                loaded_icon.pixmap.as_ref(),
                &PixmapPaint {
                    opacity: filter.transform.alpha,
                    blend_mode: BlendMode::SourceOver,
                    quality: FilterQuality::Bicubic,
                },
                Transform::from_scale(scale, scale).post_rotate_at(
                    (filter.transform.clockwise_quarter_rotations as f32) * 90.0,
                    key_size.width() as f32 / 2.0,
                    key_size.height() as f32 / 2.0,
                ),
                None,
            );
        }

        if let Some(label) = &style.label {
            if !label.is_empty() {
                context.label_renderer.borrow_mut().render(&mut pixmap, &label);
            }
        }

        if let Some(color) = style.border {
            let path = PathBuilder::from_rect(Rect::from_xywh(-1.0, -2.0, pixmap.width() as f32, pixmap.height() as f32).unwrap());

            let paint = Paint {
                shader: Shader::SolidColor(Color::from_rgba8(color.r, color.g, color.b, color.a)),
                ..Paint::default()
            };

            static STROKE: Stroke = Stroke {
                width: 15.0,
                miter_limit: 4.0,
                line_cap: LineCap::Butt,
                line_join: LineJoin::Round,
                dash: None,
            };

            pixmap.stroke_path(&path, &paint, &STROKE, Transform::identity(), None);
        }
    }

    convert_pixels_to_rgb565(pixmap.pixels(), context.buffer_endianness).freeze()
}

fn convert_pixels_to_rgb565(pixels: &[PremultipliedColorU8], endianness: Endianness) -> BytesMut {
    let pixel_count = pixels.len();

    let mut result = BytesMut::with_capacity(pixel_count * 2);

    for pixel in pixels {
        let red = pixel.red() as u16;
        let green = pixel.green() as u16;
        let blue = pixel.blue() as u16;
        let color = ((red & 0b11111000) << 8) | ((green & 0b11111100) << 3) | (blue.wrapping_shr(3));

        match endianness {
            Endianness::LittleEndian => result.put_u16_le(color),
            Endianness::BigEndian => result.put_u16(color),
        }
    }

    result
}

pub mod labels {
    use cosmic_text::{Align, Attrs, AttrsList, Buffer, BufferLine, FontSystem, Metrics, Shaping, SwashCache};
    use encode_unicode::StrExt;
    use tiny_skia::{Color, Paint, Pixmap, Rect, Shader, Transform};

    #[derive(Debug)]
    pub struct LabelRenderer {
        font_system: FontSystem,
        swash_cache: SwashCache,
        buffer: Buffer,
    }

    impl LabelRenderer {
        pub fn new(font_family: Option<&String>) -> Self {
            let mut font_system = FontSystem::new();
            if let Some(f) = font_family {
                font_system.db_mut().set_sans_serif_family(f);
            }

            let buffer = Buffer::new(&mut font_system, Metrics::new(11.0, 11.0));

            LabelRenderer {
                font_system,
                swash_cache: SwashCache::new(),
                buffer,
            }
        }

        pub fn render(&mut self, pixmap: &mut Pixmap, text: &String) {
            let mut line = BufferLine::new(text, AttrsList::new(Attrs::new()), Shaping::Advanced);
            line.set_align(Some(Align::Center));

            self.buffer.lines.clear();
            self.buffer.lines.push(line);

            const PADDING: f32 = 12.0;
            self.buffer.set_size(
                &mut self.font_system,
                pixmap.width() as f32 - PADDING * 2.0,
                pixmap.height() as f32 - PADDING * 2.0,
            );

            let font_size = if text.utf8chars().count() == 1 { 40.0 } else { 11.0 };
            self.buffer.set_metrics(&mut self.font_system, Metrics::new(font_size, font_size));

            self.buffer.shape_until_scroll(&mut self.font_system);

            self.buffer.draw(
                &mut self.font_system,
                &mut self.swash_cache,
                cosmic_text::Color::rgb(255, 255, 255),
                |x, y, w, h, color| {
                    pixmap.fill_rect(
                        Rect::from_xywh(x as f32 + PADDING, y as f32 + PADDING, w as f32, h as f32).unwrap(),
                        &Paint {
                            shader: Shader::SolidColor(Color::from_rgba8(color.r(), color.g(), color.b(), color.a())),
                            ..Paint::default()
                        },
                        Transform::identity(),
                        None,
                    );
                },
            );
        }
    }
}
