use std::cell::RefCell;
use std::collections::HashMap;
use std::path::Path;
use std::sync::Arc;
use std::thread;
use std::time::Instant;

use color_eyre::eyre::{ContextCompat, WrapErr};
use color_eyre::Result;
use enum_map::EnumMap;
use enum_ordinalize::Ordinalize;
use log::{debug, info, trace};
use rgb::RGB8;
use tiny_skia::IntSize;
use tokio::sync::broadcast;

use loupedeck_serial::characteristics::{LoupedeckButton, LoupedeckDeviceKeyGridCharacteristics};
use loupedeck_serial::commands::VibrationPattern;
use loupedeck_serial::device::LoupedeckDevice;
use loupedeck_serial::events::LoupedeckEvent;

use crate::icons::{get_used_icon_descriptors, load_icons, LoadedIconsMap};
use crate::model;
use crate::model::knob_page::KnobStyle;
use crate::model::position::{ButtonPosition, KeyPath, KeyPosition, KnobPath};
use crate::modes::key::{start_handlers, KeyEvent};
use crate::runner::graphics::labels::LabelRenderer;
use crate::runner::graphics::{render_key, GraphicsContext};
use crate::runner::state::{Key, State, StateChangeCommand};

mod graphics;
pub mod state;

pub async fn start(config_directory: &Path, config: model::config::Config) -> Result<()> {
    let config = Arc::new(config);
    info!("Discovering devices…");

    let available_devices = LoupedeckDevice::discover()?;
    let available_device = available_devices.first().wrap_err("No device connected.")?;

    info!("Connecting to the device…");
    let device = available_device.connect().wrap_err("Connecting to the device failed.")?;
    info!("Connected");

    let key_grid = &device.characteristics().key_grid;
    let used_icon_descriptors = get_used_icon_descriptors(&config);

    let start_time = Instant::now();
    info!("Loading icons…");
    let icons = load_icons(config_directory, &config.icon_packs, used_icon_descriptors, key_grid.display.dpi)?;
    info!("Finished loading {} icon(s) in {}ms", icons.len(), start_time.elapsed().as_millis());

    device.set_brightness(0.5);
    device.vibrate(VibrationPattern::RiseFall);

    let (commands_sender, commands_receiver) = flume::bounded::<StateChangeCommand>(20);
    let key_events_sender: broadcast::Sender<(KeyPath, KeyEvent)> = broadcast::Sender::new(20);

    commands_sender
        .send(StateChangeCommand::SetActivePages {
            knob_page_id: config.initial.knob_page.clone(),
            key_page_id: config.initial.key_page.clone(),
        })
        .unwrap();

    let cloned_config = Arc::clone(&config);
    let cloned_commands_sender = commands_sender.clone();
    let cloned_key_events_sender = key_events_sender.clone();
    let io_worker_thread = thread::Builder::new()
        .name("deckster IO worker".to_owned())
        .spawn(move || {
            do_io_work(
                cloned_config,
                icons,
                device,
                cloned_key_events_sender,
                cloned_commands_sender,
                commands_receiver,
            )
        })
        .wrap_err("Could not spawn the worker thread")?;

    start_handlers(
        config.key_pages_by_id.iter().flat_map(|(page_id, page)| {
            page.keys.iter().map(|(position, key)| {
                (
                    KeyPath {
                        page_id: page_id.clone(),
                        position: *position,
                    },
                    Arc::clone(key),
                )
            })
        }),
        key_events_sender,
        commands_sender,
    );

    info!("Ready.");
    io_worker_thread.join().unwrap();

    Ok(())
}

fn create_state(config: &model::config::Config) -> State {
    let key_pages_by_id: HashMap<_, _> = config
        .key_pages_by_id
        .iter()
        .map(|(id, p)| state::KeyPage {
            id: id.clone(),
            keys_by_position: p
                .keys
                .iter()
                .map(|(position, k)| Key {
                    path: KeyPath {
                        page_id: p.id.clone(),
                        position: *position,
                    },
                    base_style: k.base_style.clone(),
                    style: None,
                })
                .map(|k| (k.path.position, k))
                .collect(),
        })
        .map(|p| (p.id.clone(), p))
        .collect();

    let knob_pages_by_id: HashMap<_, _> = config
        .knob_pages_by_id
        .iter()
        .map(|(id, p)| state::KnobPage {
            id: id.clone(),
            knobs_by_position: EnumMap::from_fn(|position| {
                let knob_config = &p.knobs[position];

                state::Knob {
                    path: KnobPath {
                        page_id: p.id.clone(),
                        position,
                    },
                    base_style: knob_config.base_style.clone(),
                    style: KnobStyle::default(),
                    value: 0.0,
                }
            }),
        })
        .map(|p| (p.id.clone(), p))
        .collect();

    State {
        active_key_page_id: config.initial.key_page.clone(),
        active_knob_page_id: config.initial.knob_page.clone(),
        key_pages_by_id,
        knob_pages_by_id,
    }
}

enum IoWork {
    Event(LoupedeckEvent),
    Command(StateChangeCommand),
}

struct IoWorkerContext {
    config: Arc<model::config::Config>,
    device: LoupedeckDevice,
    state: State,
    graphics: GraphicsContext,
}

fn do_io_work(
    config: Arc<model::config::Config>,
    icons: LoadedIconsMap,
    device: LoupedeckDevice,
    key_events_sender: broadcast::Sender<(KeyPath, KeyEvent)>,
    commands_sender: flume::Sender<StateChangeCommand>,
    commands_receiver: flume::Receiver<StateChangeCommand>,
) {
    let state = create_state(&config);
    let buffer_endianness = device.characteristics().key_grid.display.endianness;
    let global_icon_filter_by_pack_id = config
        .icon_packs
        .iter()
        .filter_map(|(i, p)| p.global_filter.clone().map(|f| (i.clone(), f)))
        .collect();

    let label_renderer = RefCell::new(LabelRenderer::new(config.label_font_family.as_ref()));

    let device_events_receiver = device.events();

    let mut context = IoWorkerContext {
        config,
        device,
        state,
        graphics: GraphicsContext {
            loaded_icons: icons,
            buffer_endianness,
            label_renderer,
            global_icon_filter_by_pack_id,
        },
    };

    loop {
        let a = flume::Selector::new()
            .recv(&device_events_receiver, |e| IoWork::Event(e.unwrap()))
            .recv(&commands_receiver, |c| IoWork::Command(c.unwrap()))
            .wait();

        match a {
            IoWork::Event(event) => {
                if !handle_event(&context, &commands_sender, &key_events_sender, event) {
                    break;
                }
            }
            IoWork::Command(command) => handle_command(&mut context, command),
        }
    }
}

fn handle_event(
    context: &IoWorkerContext,
    commands_sender: &flume::Sender<StateChangeCommand>,
    key_events_sender: &broadcast::Sender<(KeyPath, KeyEvent)>,
    event: LoupedeckEvent,
) -> bool {
    trace!("Handling event: {:?}", &event);

    let send_key_event = |path: KeyPath, event: KeyEvent| {
        trace!("Sending key event ({}): {:?}", &path, &event);
        key_events_sender.send((path, event)).unwrap();
    };

    match event {
        LoupedeckEvent::Disconnected => return false,
        LoupedeckEvent::ButtonDown { button } => {
            let position = ButtonPosition::of(&button);
            let button_config = &context.config.buttons[position];

            commands_sender
                .send(StateChangeCommand::SetActivePages {
                    key_page_id: button_config.key_page.as_ref().unwrap_or(&context.state.active_key_page_id).clone(),
                    knob_page_id: button_config.knob_page.as_ref().unwrap_or(&context.state.active_knob_page_id).clone(),
                })
                .unwrap()
        }
        LoupedeckEvent::Touch { x, y, is_end, .. } => {
            if is_end {
                let characteristics = context.device.characteristics();
                let display = characteristics.get_display_at_coordinates(x, y);

                if let Some(display) = display {
                    if display.name == characteristics.key_grid.display.name {
                        let key_index = characteristics.key_grid.get_key_at_global_coordinates(x, y);
                        if let Some(key_index) = key_index {
                            let position = KeyPosition {
                                x: (key_index % characteristics.key_grid.columns) as u16 + 1,
                                y: (key_index / characteristics.key_grid.columns) as u16 + 1,
                            };

                            let path = KeyPath {
                                page_id: context.state.active_key_page_id.clone(),
                                position,
                            };

                            send_key_event(path, KeyEvent::Press);
                        }
                    }
                }
            }
        }
        _ => {}
    }

    true
}

fn handle_command(context: &mut IoWorkerContext, command: StateChangeCommand) {
    debug!("Handling command: {:?}", &command);

    match command {
        StateChangeCommand::SetActivePages { key_page_id, knob_page_id } => {
            context.state.active_key_page_id = key_page_id;
            context.state.active_knob_page_id = knob_page_id;

            for button in LoupedeckButton::VARIANTS {
                let position = ButtonPosition::of(button);

                context.device.set_button_color(*button, get_correct_button_color(context, position)).unwrap();
            }

            let key_grid = &context.device.characteristics().key_grid;
            for index in 0..(key_grid.rows * key_grid.columns) {
                draw_key_at_index(context, index);
            }

            context.device.refresh_display(&key_grid.display).unwrap();
        }
        StateChangeCommand::SetKeyStyle { path, value } => {
            context.state.mutate_key_for_command("SetKeyStyle", &path, |k| {
                k.style = value;
            });

            draw_key_at_path_if_visible(context, path);
            context.device.refresh_display(&context.device.characteristics().key_grid.display).unwrap();
        }
    }
}

// active -> config.active_button_color
// no actions defined -> #000000
// inactive -> config.inactive_button_color
fn get_correct_button_color(context: &IoWorkerContext, button_position: ButtonPosition) -> RGB8 {
    let button_config = &context.config.buttons[button_position];

    if let Some(key_page) = &button_config.key_page {
        if key_page == &context.state.active_key_page_id {
            if let Some(knob_page) = &button_config.knob_page {
                if knob_page == &context.state.active_knob_page_id {
                    return context.config.active_button_color.into();
                }
            }
        }
    } else if button_config.knob_page.is_none() {
        return RGB8::new(0, 0, 0);
    }

    context.config.inactive_button_color.into()
}

fn get_key_index_for_position(key_grid: &LoupedeckDeviceKeyGridCharacteristics, position: KeyPosition) -> Option<u8> {
    if (position.x - 1) >= key_grid.columns as u16 || (position.y - 1) >= key_grid.rows as u16 {
        None
    } else {
        let x = (position.x - 1) as u8;
        let y = (position.y - 1) as u8;
        Some(y * key_grid.columns + x)
    }
}

fn get_key_position_for_index(key_grid: &LoupedeckDeviceKeyGridCharacteristics, index: u8) -> KeyPosition {
    let x = index % key_grid.columns;
    let y = index / key_grid.columns;

    KeyPosition {
        x: (x + 1) as u16,
        y: (y + 1) as u16,
    }
}

fn draw_key(context: &IoWorkerContext, index: u8, key: Option<&Key>) {
    let key_grid = &context.device.characteristics().key_grid;
    let (x, y, w, h) = key_grid.get_local_key_rect_xywh(index).unwrap();

    let p = render_key(&context.graphics, IntSize::from_wh(w as u32, h as u32).unwrap(), key);

    context.device.replace_framebuffer_area_raw(&key_grid.display, x, y, w, h, p).unwrap();
}

fn draw_key_at_index(context: &IoWorkerContext, index: u8) {
    let position = get_key_position_for_index(&context.device.characteristics().key_grid, index);

    draw_key(context, index, context.state.active_key_page().keys_by_position.get(&position));
}

fn draw_key_at_position_if_visible(context: &IoWorkerContext, position: KeyPosition) {
    let index = get_key_index_for_position(&context.device.characteristics().key_grid, position);

    if let Some(index) = index {
        draw_key(context, index, context.state.active_key_page().keys_by_position.get(&position));
    }
}

fn draw_key_at_path_if_visible(context: &IoWorkerContext, path: KeyPath) {
    if context.state.active_key_page_id == path.page_id {
        draw_key_at_position_if_visible(context, path.position);
    }
}
