use std::collections::HashMap;

use enum_map::EnumMap;
use log::error;
use serde::{Deserialize, Serialize};

use crate::model::key_page::KeyStyle;
use crate::model::knob_page::KnobStyle;
use crate::model::position::{KeyPath, KeyPosition, KnobPath, KnobPosition};

#[derive(Debug)]
pub struct State {
    pub active_key_page_id: String,
    pub active_knob_page_id: String,
    pub key_pages_by_id: HashMap<String, KeyPage>,
    pub knob_pages_by_id: HashMap<String, KnobPage>,
}

impl State {
    pub fn mutate_key_for_command<R>(&mut self, command_name: &'static str, path: &KeyPath, mutator: impl FnOnce(&mut Key) -> R) -> Option<R> {
        match self.key_pages_by_id.get_mut(&path.page_id) {
            None => error!("Received {} command with invalid path.page_id: {}", command_name, &path.page_id),
            Some(key_page) => match key_page.keys_by_position.get_mut(&path.position) {
                None => error!("Received {} command with invalid path.position: {}", command_name, &path.position),
                Some(key) => return Some(mutator(key)),
            },
        }

        None
    }

    pub fn get_key(&self, path: &KeyPath) -> &Key {
        &self.key_pages_by_id[&path.page_id].keys_by_position[&path.position]
    }

    pub fn active_key_page(&self) -> &KeyPage {
        &self.key_pages_by_id[&self.active_key_page_id]
    }

    pub fn active_knob_page(&self) -> &KnobPage {
        &self.knob_pages_by_id[&self.active_knob_page_id]
    }
}

#[derive(Debug)]
pub struct KeyPage {
    pub id: String,
    pub keys_by_position: HashMap<KeyPosition, Key>,
}

#[derive(Debug)]
pub struct KnobPage {
    pub id: String,
    pub knobs_by_position: EnumMap<KnobPosition, Knob>,
}

#[derive(Debug)]
pub struct Key {
    pub path: KeyPath,
    pub base_style: KeyStyle,
    pub style: Option<KeyStyle>,
}

#[derive(Debug)]
pub struct Knob {
    pub path: KnobPath,
    pub base_style: KnobStyle,
    pub style: KnobStyle,
    pub value: f32,
}

#[derive(Debug, PartialEq, Clone, Serialize, Deserialize)]
#[allow(clippy::enum_variant_names)]
pub enum StateChangeCommand {
    SetActivePages { key_page_id: String, knob_page_id: String },
    SetKeyStyle { path: KeyPath, value: Option<KeyStyle> },
}
